// This is -*- C++ -*-
// $Id: coherent1.cc,v 1.5 1998/08/19 20:53:56 trow Exp $

/* coherent1.cc
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Jon Trowbridge <trow@emccta.com>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <iostream.h>
#include <math.h>
#include <algo.h>
#include "Random.h"
#include "DataSet.h"


// Manually check that a DataSet is in a coherent state by redoing
// certain calculations.

bool
eq(const DataSet& ds, double x, double y)
{
  return fabs(x-y)<ds.epsilon();
}

void
coherent(const DataSet& ds)
{
  cout << "Checking coherency of " << ds.label() << endl;

  // FIXME: check special case of size == 0
  if (ds.size() == 0) return;

  cout << "Recalculating descriptive stats..." << endl;
  double min=ds.data(0), max=ds.data(0), mean=ds.data(0);
  for(size_t i=1; i<ds.size(); ++i) {
    double x = ds.data(i);
    if (x < min) min = x;
    if (x > max) max = x;
    mean += x;
  }
   mean /= ds.size();
  if (! eq(ds, min, ds.min()) )
    cout << "*** min() is " << ds.min() << ", should be " << min << endl;
  if (! eq(ds, max, ds.max()) )
    cout << "*** max() is " << ds.max() << ", should be " << max << endl;
  if (! eq(ds, mean, ds.mean()) ) {
    cout << "*** mean() is " << ds.mean() << ", should be " << mean << endl;
  }
  cout << "Done" << endl;

  // check things about sortedness

  cout << "Checking sort status..." << endl;
  double smean=ds.sorted_data(0), slast=ds.sorted_data(0), last=ds.data(0);
  bool unsorted=false;
  int sortdir = 1;
  for(size_t i=1; i<ds.size(); ++i) {
    double sx = ds.sorted_data(i);
    if (sx < slast) unsorted = true;
    smean += sx;

    if (sortdir) {
      double x = ds.data(i);
      if (i == 1) {
	sortdir = last <= x ? 1 : -1;
      } else {
	if (sortdir ==  1 && last > x) sortdir = 0;
	if (sortdir == -1 && last < x) sortdir = 0;
      }
      last = x;
    }
  }

  smean /= ds.size();

  if (unsorted)
    cout << "*** sorted_data() refers to unsorted data!" << endl;
  if (sortdir != ds.sortedness()) 
    cout << "*** sortedness() is " << ds.sortedness() << ", should be " << sortdir << endl;

  if (! eq(ds, mean, smean))
    cout << "*** mean of sorted_data() != mean of data()" << endl;
  cout << "Done" << endl;

  // make sure that the sorted data and the data are really the same...
  cout << "Matching data and sorted data..." << endl;
  double* d = new double[ds.size()];
  memcpy(d,ds.data(),sizeof(double)*ds.size());
  sort(d,d+ds.size(),less<double>());
  int mismatch=0;
  for(size_t i=0; i<ds.size(); ++i)
    if (d[i] != ds.sorted_data(i)) ++mismatch;
  if (mismatch) {
    cout << "*** mismatch between data() and sorted_data(): " << mismatch << " out of " << ds.size() << " elements disagree" << endl;
    for(size_t i=0; i<ds.size(); ++i) 
      cout << "    " << d[i] << " " << ds.sorted_data(i) << endl;
  }
  delete [] d;
  cout << "Done" << endl;

  cout << endl;
}

double foo_transform(double x)
{
  return sin(3.14159*x);
}

main()
{
  DataSet ds;
  Random rnd;

  for(int i=0;i<1000;++i) 
    ds.add(rnd.random_uniform(-1,1));


  for(;;) {
    int test = rnd.random_integer(1,8);

    if (test == 1) {
      ds.sort();
      ds.set_label("forward-sorted random data");
    } else if (test == 2) {
      ds.sort(-1);
      ds.set_label("backward-sorted random data");
    } else if (test == 3) {
      ds.scramble();
      ds.set_label("scrambled random data");
    } else if (test == 4) {
      ds.apply(foo_transform);
      ds.set_label("transformed random data");
    } else if (test == 5) {
      ds.rank();
      ds.set_label("ranked random data");
    } else if (test == 6) {
      ds.reverse();
      ds.set_label("reversed random data");
    } else if (test == 7) {
      for(int i=0; i<2; ++i) 
	ds.add(rnd.random_uniform(-1,1));
      ds.set_label("extended random data");
    } else if (test == 8) {
      ds.linear_transform(rnd.random_normal(0.0,2.0), rnd.random_normal());
      ds.set_label("linearly transformed random data");
    }

    coherent(ds);
  }

}


// $Id: coherent1.cc,v 1.5 1998/08/19 20:53:56 trow Exp $
