// This is -*- C++ -*-
// $Id: DataSet.cc,v 1.9 1998/08/20 22:03:44 trow Exp $

/* 
 * DataSet.cc
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Jon Trowbridge <trow@emccta.com>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <string.h>
#include <stdlib.h>
#include <iostream.h>
#include <algo.h>
#include "Random.h"
#include "DataSet.h"

DataSet::DataSet() : data_(0), sdata_(0), epsilon_(1e-8)
{
  clear();
}

const double*
DataSet::sorted_data() const
{
  if (size() == 0 || data_ == 0) return 0; 
  if (sdata_) return sdata_;

  double* s=0;
  if (sortedness() == 1)
    s = data_;
  else {
    s = new double[size()];
    memcpy(s, data_, sizeof(double)*size());
      
    if (sortedness() == -1) {
      // reverse the data
      for(size_t i=0; i<size()/2; ++i) {
	double t = s[i];
	s[i] = s[size()-1-i];
	s[size()-1-i] = t;
      }
    } else {
      // sort the data
      ::sort(s, s+size(), less<double>());
    }
  }
    
  // Sometimes you just have to fool the compiler...
  ((DataSet*)this)->sdata_ = s;

  return (const double*)s;
}

double
DataSet::percentile(double p) const
{
  const double* s = sorted_data();
  double t = p*(size()-1);
  int i = (int)t;
  return s ? (i+1-t)*s[i] + (t-i)*s[i+1] : NAN;
}

double
DataSet::moment(unsigned k, double x) const
{
  if (size() == 0) return NAN;

  double sum=0;
  for(size_t i=0; i<size(); ++i)
    sum += pow(data_[i]-x,k);
  return sum/size();
}

double
DataSet::gmean() const
{
  if (size() == 0 || !is_positive(min())) return NAN;
  double logsum=0;
  for(size_t i=0; i<size(); ++i)
    logsum += log(data_[i]);
  return exp(logsum/size());
}

double
DataSet::hmean() const
{
  if (size() == 0) return NAN;
  double invsum=0;
  for(size_t i=0; i<size(); ++i) {
    if (is_zero(data_[i])) return NAN;
    invsum += 1/data_[i];
  }
  return is_zero(invsum) ? NAN : size()/invsum;
}

double
DataSet::rms() const
{
  // "My work on free software is motivated by an idealistic goal: spreading
  // freedom and cooperation.  I want to encourage free software to spread,
  // replacing proprietary software which forbids cooperation, and thus
  // make our society better." -rms
  if (size() == 0) return NAN;
  double sum=0;
  for(size_t i=0; i<size(); ++i)
    sum += data_[i]*data_[i];
  return sqrt(sum/size());
}

double
DataSet::meandev() const
{
  if (size() == 0) return NAN;
  double sum=0, m=mean();
  for(size_t i=0; i<size(); ++i)
    sum += fabs(data_[i]-m);
  return sum/size();
}

double
DataSet::meddev() const
{
  // This is expensive, and it duplicates code.  And it could be optimized.
  // Yipee!
  if (size() == 0) return NAN;
  double m=med();
  double* d = new double[size()];
  for(size_t i=0; i<size(); ++i)
    d[i] = fabs(data_[i]-m);
  ::sort(d,d+size(),less<double>());
  double md = size()%2 ? d[size()/2] : (d[size()/2-1]+d[size()/2])/2;
  delete [] d;
  return md;
}

double
DataSet::kurtosis() const
{
  if (size() == 0) return NAN;
  double m=mean(), mom2=0, mom4=0;
  for(size_t i=0; i<size(); ++i) {
    double x = data_[i]-m;
    x *= x;
    mom2 += x;
    mom4 += x*x;
  }
  return mom4/(mom2*mom2);
}

double
DataSet::skewness() const
{
  if (size() == 0) return NAN;
  double m=mean(), mom2=0, mom3=0;
  for(size_t i=0; i<size(); ++i) {
    double x = data_[i]-m;
    mom2 += x*x;
    mom3 += x*x*x;
  }
  return mom3/pow(mom2,1.5);
}

double
DataSet::covar(const DataSet& ds) const
{
  if (size() != ds.size()) return NAN;
  double m1=mean(), m2=ds.mean();
  double sum=0;
  for(size_t i=0; i<size(); ++i)
    sum += (data_[i]-m1)*(ds.data_[i]-m2);
  return sum / size();
}

size_t
DataSet::greater_than(double x)
{
  // FIXME: optimize for the case where the data is sorted
  size_t N=0;
  for(size_t i=0; i<size(); ++i)
    if (data_[i] > x) ++N;
  return N;
}

size_t
DataSet::less_than(double x)
{
  // FIXME: optimize for the case where the data is sorted
  size_t N=0;
  for(size_t i=0; i<size(); ++i)
    if (data_[i] < x) ++N;
  return N;
}

size_t
DataSet::between(double a, double b)
{
  // FIXME: optimize for the case where the data is sorted
  size_t N=0;
  for(size_t i=0; i<size(); ++i)
    if (a <= data_[i] && data_[i] <= b) ++N;
  return N;
}

void
DataSet::reserve(size_t R)
{
  if (R > slots_) {
    bool keep_sdata = (data_ && sdata_ == data_);
    double* t = new double[R];
    if (data_ && size_)
      memcpy(t, data_, sizeof(double)*size_);
    delete [] data_;
    data_ = t;
    slots_ = R;
    if (keep_sdata)
      sdata_ = data_;
    else {
      delete [] sdata_;
      sdata_ = 0;
    }
  }
}

void
DataSet::add(double x)
{
  if (size_ == slots_) {
    reserve(slots_ == 0 ? 256 : 2*slots_); // nice round # as default
  }

  // update statistics /////

  if (size_ == 0 || x < min_)
    min_ = x;
  if (size_ == 0 || x > max_)
    max_ = x;
  
  sum_ += x;
  if (size_ > 0) {
    sumsq_ += (x-sum_/(size_+1))*(x-(sum_-x)/size_); // cool, huh?

    // Update sortedness flag
    if (sortedness_) {
      // Fixme: this does the wrong thing if the data is down-sorted
      // but the first two elements are ==.
      if (size_ == 1)
	sortedness_ = (data_[0] <= x) ? 1 : -1;
      else if ( (sortedness_ ==  1 && data_[size_-1] > x) ||
		(sortedness_ == -1 && data_[size_-1] < x))
	sortedness_ = 0;
    }
  }

  // FIXME: this could be optimized
  if (sdata_ != data_)
    delete [] sdata_;
  sdata_ = 0;

  //////////////////////////

  data_[size_++] = x;
}

void
DataSet::clear()
{
  size_ = slots_ = 0;
  if (data_ != sdata_)
    delete [] sdata_;
  delete [] data_;
  data_ = sdata_ = 0;
  min_ = max_ = NAN;
  sum_ = sumsq_ = 0;
  sortedness_ = 1;
}

void
DataSet::copy(const DataSet& ds)
{
  clear();
  reserve(ds.slots_);
  size_ = ds.size_;
  min_ = ds.min_;
  max_ = ds.max_;
  sum_ = ds.sum_;
  sumsq_ = ds.sumsq_;
  sortedness_ = ds.sortedness_;
  memcpy(data_, ds.data_, sizeof(double)*size_);
}

void
DataSet::append(const DataSet& ds)
{
  // clearly, this could be optimized...
  for(size_t i=0; i<ds.size(); ++i)
    add(ds.data(i));
}

void
DataSet::apply(double (*fn)(double))
{
  for(size_t i=0; i<size(); ++i)
    data_[i] = fn(data_[i]);
  recalc();
}

void
DataSet::linear_transform(double a, double b)
{
  for(size_t i=0; i<size(); ++i) 
    data_[i] = a*data_[i]+b;

  
  double m1 = a*min_+b, m2 = a*max_+b;
  min_ = m1<m2 ? m1 : m2;
  max_ = m1<m2 ? m2 : m1;

  sum_ = a*sum_+b*size();
  sumsq_ *= a*a;
  if (a<0) sortedness_ = -sortedness_;

  if (sdata_) {
    if (sdata_ == data_) {

      if (a<0) {
	sdata_ = 0;
      }

    } else {

      // linearly transform our sdata buffer
      for(size_t i=0; i<size(); ++i)
	sdata_[i] = a*sdata_[i]+b;

      // if a<0, reverse the sdata to preserve sortedness
      if (a<0) {
	for(size_t i=0; i<size()/2; ++i) {
	  double t=sdata_[i];
	  sdata_[i]=sdata_[size()-1-i];
	  sdata_[size()-1-i]=t;
	}
      }

    }
  }
}

void
DataSet::log_transform()
{
  if (min() <= 0) {
    clear();
    return;
  }

  for (size_t i=0; i<size(); ++i)
    data_[i] = log(data_[i]);

  double* saved_sdata = 0;
  if (sdata_ && sdata_ != data_) {
    for(size_t i=0; i<size(); ++i)
      sdata_[i] = log(sdata_[i]);
    
    // This is hack that lets us call recalc() but still keep our
    // sorted data, if we have it.
    saved_sdata = sdata_;
    sdata_ = 0;
  }
  recalc();
  sdata_ = saved_sdata;
}

void
DataSet::reverse()
{
  if (sdata_ == data_)
    sdata_ = 0;

  for(size_t i=0; i<size_/2; ++i) {
    double t = data_[i];
    data_[i] = data_[size_-i-1];
    data_[size_-1-i] = t;
  }
  sortedness_ = -sortedness_;
}

void
DataSet::sort(int dir)
{
  // If we are already sorted properly, ignore
  if ((sortedness_ == 1 && dir >= 0) || (sortedness_ == -1 && dir < 0))
    return;

  // If we have sorted data already, use it!
  if (dir >= 0 && sdata_ && sdata_ != data_) {
    memcpy(data_, sdata_, sizeof(double)*size());
    sortedness_ = 1;
    return;
  }

  // If we are being asked to "opposite-sort" already-sorted data, do so
  // via reverse()
  if ((dir >= 0 && sortedness_ == -1) || (dir < 0 && sortedness_ == 1)) {
    reverse();
    sortedness_ = dir >= 0 ? 1 : -1;
  } else if (dir >= 0) {
    // Otherwise there is no avoiding it, so just sort the data...
    ::sort(data_, data_+size_, less<double>());
    sortedness_ = 1;
  } else {
    ::sort(data_, data_+size_, greater<double>());
    sortedness_ = -1;
  }
  
  // Alias sdata_ and data_, if applicable.
  if (sortedness_ == 1 && sdata_ != data_) {
    delete [] sdata_;
    sdata_ = data_;
  }
}

void
DataSet::scramble()
{
  if (size() == 0) return;

  // If we were sorted before, copy the sorted data so we don't lose it
  if (sdata_ == data_) {
    sdata_ = new double[size()];
    memcpy(sdata_, data_, sizeof(double)*size());
  }

  for(size_t i=0; i<size_-1; ++i) {
    size_t j = rnd.random_integer(i,size_-1);
    if (i != j) {
      double t = data_[i];
      data_[i] = data_[j];
      data_[j] = t;
    }
  }

  if (data_ == sdata_)
    sdata_ = 0;

  sortedness_ = 0;
}

void
DataSet::rank()
{
  pair<double,int>* rankdata = new pair<double,int>[size_];

  for(size_t i=0; i<size(); ++i) {
    rankdata[i].first = data_[i];
    rankdata[i].second = i;
  }

  ::sort(rankdata, rankdata+size_, less< pair<double,int> >());

  // set ranks, handling ties by the usual convention
  unsigned count=1;
  for(size_t i=0; i<size_; /* increment down below */ ) {
    size_t j=i+1;
    unsigned count2=0;
    while (j < size_ && are_equal(rankdata[i].first, rankdata[j].first)) {
      ++j;
      ++count2;
    }
    while (i != j)
      rankdata[i++].first = (2*count + count2)/2.0; // we increment i here
    count += 1+count2;
  }

  for(size_t i=0; i<size_; ++i)
    data_[rankdata[i].second] = rankdata[i].first;

  recalc(); // This could be optimized...

  delete [] rankdata;
}

// There is some code duplication between this and add(), but calling
// recalc() should be a lot more efficient than just repeatedly add()-ing
// the numbers if you've been fiddling directly with data_.
void
DataSet::recalc()
{
  min_ = max_ = NAN;
  sum_ = sumsq_ = 0;
  sortedness_ = 1;
  if (size_ == 0) return;

  min_ = max_ = data_[0];
  for (size_t i=0; i<size_; ++i) {
    double x = data_[i];
    if (min_ > x) min_ = x;
    if (max_ < x) max_ = x;
    sum_ += x;
    if (i>0) {
      sumsq_ += (x-sum_/(size_+1))*(x-(sum_-x)/size_);
      if (sortedness_) {
	// Fixme: this does the wrong thing if the data is down-sorted
	// but the first two elements are ==.
	if (i == 1) 
	  sortedness_ = (data_[0] <= x) ? 1 : -1;
	else if ( (sortedness_ ==  1 && data_[i-1] > x) ||
		  (sortedness_ == -1 && data_[i-1] < x))
	  sortedness_ = 0;
      }
    }
  }

  if (data_ != sdata_)
    delete [] sdata_;
  sdata_ = 0;
}

// $Id: DataSet.cc,v 1.9 1998/08/20 22:03:44 trow Exp $

