// This is -*- C++ -*-
// $Id: statfn.cc,v 1.5 1998/08/21 17:21:48 trow Exp $


/* statfn.cc
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Jon Trowbridge <trow@emccta.com>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <iostream.h>
#include <math.h>
#include <nan.h>
#include "statfn.h"

double
normal_cdf(double x)
{
  x /= M_SQRT2;
  return 0.5 * (x >= 0 ? 1 + erf(x) : erfc(-x));
}

double
inverse_normal_cdf(double x)
{
  int sign;
  if (x < 0.5) {
    sign = -1;
  } else {
    sign = 1;
    x = 1-x;
  }

  if (x < 1e-20) return sign*10;

  x = sqrt(-2*log(x));    
  
  const double p0 = -0.322232431088;
  const double p1 = -1.0;
  const double p2 = -0.342242088547;
  const double p3 = -0.0204231210245;
  const double p4 = -0.453642210148e-4;

  const double q0 = 0.0993484626060;
  const double q1 = 0.588581570495;
  const double q2 = 0.531103462366;
  const double q3 = 0.103537752850;
  const double q4 = 0.38560700634e-2;

  double num = (((x*p4+p3)*x+p2)*x+p1)*x+p0;
  double den = (((x*q4+q3)*x+q2)*x+q1)*x+q0;
  return sign*(x+num/den);
}

double
log_gamma(double x)
{
  double sum = 1.000000000190015
    + 76.18009172947146 / (x+1)
    - 86.50532032941677 / (x+2)
    + 24.01409824083091 / (x+3)
    - 1.231739572450155 / (x+4)
    + 0.1208650973866179e-2 / (x+5)
    - 0.5395239384953e-5 / (x+6);

  return log(2.5066282746310005 * sum / x) - x - 5.5 + (x + 0.5)*log(x + 5.5);
}

double
incomplete_gamma(double a,double x)
{
  const double EPSILON = 1e-15;

  if (x < 0.0 || a <= 0.0) return NAN;
  if (x < a+1) { /* use series */
    double term=1/a, sum=1/a;
    for (int i=1; ; ++i) {
      term *= x/(a+i);
      sum += term;
      if (fabs(term) < fabs(sum)*EPSILON)
        return sum*exp(-x+a*log(x)-log_gamma(a));
    }
  } else { /* use continued fraction, via Lentz's algorithm */
    const double TINY = 1e-30;
    double b=x+1-a, c=1/TINY, d=1/b, f=d;
    for(int i=1; ; ++i) {
      double aa = (a-i)*i;
      b += 2;
      c = b+aa/c;
      d = aa*d+b;
      if (fabs(d) < TINY) d=TINY;
      if (fabs(c) < TINY) c=TINY;
      d=1/d;
      double term=c*d;
      f *= term;
      if (fabs(term-1) < EPSILON)
	return 1-exp(-x+a*log(x)-log_gamma(a))*f;
    }
  }
  return -1; /* to suppress compiler warning messages */
}

double
incomplete_beta(double a,double b,double x)
{
  if (x < 0 || x > 1) return NAN;
  if (x == 0 || x == 1) return x;

  double factor =
    exp(log_gamma(a+b)-log_gamma(a)-log_gamma(b)+a*log(x)+b*log(1-x));

  bool flip = false;
  if (x < (a+1)/(a+b+2)) {
    flip = true;
    double t=a;
    a=b;
    b=t;
    x=1-x;
  }

  const double EPSILON = 1e-15, TINY = 1e-30;
  double c = 1.0, d = 1.0 - (a+b)*x/(a+1);
  if (fabs(d) < TINY) d=TINY;
  d=1/d;
  double f=d;

  for(int m=1; ; ++m) {
    // evaluate the c.f. by Lentz, doing an odd & even terms in each iteration
    double m2=2*m;
    double aa=m*(b-m)*x/((a+m2-1)*(a+m2));
    c=1+aa/c;
    d=1+aa*d;
    if (fabs(c)<TINY) c=TINY;
    if (fabs(d)<TINY) d=TINY;
    d=1/d;
    f *= d*c;

    aa=-(a+m)*(a+b+m)*x/((a+m2)*(a+m2+1));
    c=1+aa/c;
    d=1+aa*d;
    if (fabs(c)<TINY) c=TINY;
    if (fabs(d)<TINY) d=TINY;
    d=1/d;
    double z = d*c;
    f *= z;
    if (fabs(z-1)<EPSILON) {
      double res = f*factor/a;
      return flip ? 1-res : res;
    }
  }
}

//////////////////////////////////////////////////////////////////////////

double
beta_function(double z, double w)
{
  return exp(log_gamma(z)+log_gamma(w)-log_gamma(z+w));
}

double
log_factorial(double x)
{
  return x >= 0 ? log_gamma(1+x) : NAN;
}

double
poisson_cdf(double k, double x)
{
  return x <= 0 || k < 1 ? NAN : 1 - incomplete_gamma(k,x);
}

double
chi_square_cdf(double chi2, double df)
{
  return chi2 <= 0 ? 0 : incomplete_gamma(df/2,chi2/2);
}

double
students_cdf(double t, double df)
{
  double b = 1 - incomplete_beta(df/2, 0.5, df/(df+t*t));
  return 0.5 + (t >= 0 ? b/2 : -b/2);
}

double
F_cdf(double F, double df1, double df2)
{
  return incomplete_beta(df2/2, df1/2, df2/(df2+df1*F));
}




// $Id: statfn.cc,v 1.5 1998/08/21 17:21:48 trow Exp $
